/*
 * Copyright (C) 2003-2007 
 * Nick Velloff <nick.velloff@gmail.com>
 * Derrick Grigg <dgrigg@rogers.com>
 * Sean Voisen <sean@voisen.org>
 * Sean Treadway <seant@oncotype.dk>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 *
 * 
 * Copyright 2009 Jozsef Vass
 *
 * Copyright 2010 Folkert Niewijk (Chatventure.com)
 * 
 */
	 
package org.igniterealtime.xiff.core
{	 
	import adobe.utils.CustomActions;
	import com.hurlant.crypto.tls.TLSEvent;
	import com.hurlant.crypto.tls.TLSSocket;
	import flash.net.Socket;
	import flash.utils.ByteArray;
	import org.igniterealtime.xiff.core.XMPPConnection;
	import org.igniterealtime.xiff.events.DisconnectionEvent;
	import org.igniterealtime.xiff.events.IncomingDataEvent;
	import org.igniterealtime.xiff.events.OutgoingDataEvent;
	
	import flash.events.Event;
	import flash.events.IOErrorEvent;
	import flash.events.ProgressEvent;
	import flash.events.SecurityErrorEvent;
	import flash.xml.XMLDocument;
	import flash.xml.XMLNode;
	import flash.errors.IOError;
	
	public class XMPPTLSConnection extends XMPPConnection
	{
		
		protected var tlsSocket:TLSSocket;
		protected var _tlsRequired:Boolean = false;
		protected var _tlsEnabled:Boolean = false;
		
		public function XMPPTLSConnection()
		{
			super();
			
		}
		
		private function configureTLSSocket():void {
			tlsSocket = new TLSSocket();
			
			tlsSocket.addEventListener(ProgressEvent.SOCKET_DATA, socketDataReceived);
	        tlsSocket.addEventListener(Event.CLOSE, socketClosed);
	        tlsSocket.addEventListener(Event.CONNECT, socketConnected);
	        tlsSocket.addEventListener(IOErrorEvent.IO_ERROR, onIOError);
	        tlsSocket.addEventListener(SecurityErrorEvent.SECURITY_ERROR, onSecurityError);
			
	    }
		
		override public function connect( streamType:uint = 0 ):Boolean
		{
			if (_tlsEnabled)
			{
				removeTLSSocketEventListeners();
				_tlsEnabled = false;
			}
			
			createSocket();
			_streamType = streamType;

			active = false;
			loggedIn = false;

			chooseStreamTags(streamType);

			socket.connect( server, port);
			return true;
		}

		override public function disconnect():void
		{
			if( isActive() ) {
				sendXML( closingStreamTag );
				try
				{
					if (_tlsEnabled)
					{
						this.removeTLSSocketEventListeners();
						tlsSocket.close();
					}
					else
					{
						this.removeSocketEventListeners();
						socket.close();
					}
				}
				catch (e:IOError)
				{
					dispatchError(e);
				}
				active = false;
				loggedIn = false;
				var event:DisconnectionEvent = new DisconnectionEvent();
				dispatchEvent(event);
			}
		}
		
		protected function removeSocketEventListeners():void
		{
			socket.removeEventListener(Event.CLOSE, socketClosed);
			socket.removeEventListener(Event.CONNECT, socketConnected);
			socket.removeEventListener(IOErrorEvent.IO_ERROR, onIOError);
			socket.removeEventListener(ProgressEvent.SOCKET_DATA, socketDataReceived);
			socket.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, onSecurityError);
		}
		
		protected function removeTLSSocketEventListeners():void
		{
			tlsSocket.removeEventListener(ProgressEvent.SOCKET_DATA, socketDataReceived);
	        tlsSocket.removeEventListener(Event.CLOSE, socketClosed);
	        tlsSocket.removeEventListener(Event.CONNECT, socketConnected);
	        tlsSocket.removeEventListener(IOErrorEvent.IO_ERROR, onIOError);
	        tlsSocket.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, onSecurityError);
		}
		
		protected override function handleStreamFeatures( node:XMLNode ):void
		{
			if (!loggedIn)
			{
				for each(var feature:XMLNode in node.childNodes)
				{
					if (feature.nodeName == "starttls")
					{
						handleStreamTLS(feature);
					}
					else if (feature.nodeName == "mechanisms")
					{
						configureAuthMechanisms(feature);
					}
					else if (feature.nodeName == "compression")
					{
						// zlib is the most common and the one which is required to be implemented.
						if (_compress)
						{
							configureStreamCompression();
						}
					}
				}

				// Authenticate only after the possible compression
				// Authenticate only after the possible TLS encoding negotiation.
				if (((_tlsEnabled && _tlsRequired) || !_tlsRequired) && ((compress && compressionNegotiated) || !compress))
				{
					// TODO: Why is the username required here but it is not used at the backend?
					if (useAnonymousLogin || (username != null && username.length > 0))
					{
						beginAuthentication();
					}
					else
					{
						getRegistrationFields();
					}
				}
			}
			else
			{
				bindConnection();
			}
		}
		
		protected override function handleStreamTLS( node:XMLNode ):void
		{
			//if the user turned it on on the Client. If you are here the server offers TLS.
			if (_tlsRequired)
			{
				this.sendXML("<starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls' />");
			}
			else if (node.firstChild && node.firstChild.nodeName == "required") //if the user did not turn it on but the server needs it
			{
				_tlsRequired = true;
				this.sendXML("<starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls' />");
			}
		}
		
		protected override function handleNodeType(node:XMLNode):void
		{
			var nodeName:String = node.nodeName.toLowerCase();
			//<starttls xmlns='urn:ietf:params:xml:ns:xmpp-tls' />
			//<proceed xmlns='urn:ietf:params:xml:ns:xmpp-tls'  />
			switch( nodeName )
			{
				case "proceed":	
						startTLS();
					break;
				default:
					super.handleNodeType(node);
					break;
			}
		}
		
		protected function startTLS():void
		{
			this.removeSocketEventListeners();
			configureTLSSocket();
			tlsSocket.startTLS(socket, this.server);
			socket = tlsSocket;
			_tlsEnabled = true;
			
			//we have tls. The socket is now wrapped as a tlsSocket
			//now continue registering/logging in/binding connection.
			if (!loggedIn)
			{
				// Authenticate only after the possible compression
				if (((compress && compressionNegotiated) || !compress))
				{
					// TODO: Why is the username required here but it is not used at the backend?
					if (useAnonymousLogin || (username != null && username.length > 0))
					{
						beginAuthentication();
					}
					else
					{
						getRegistrationFields();
					}
				}
			}
			else
			{
				bindConnection();
			}
		}
		
		public function set tls( tlsBool:Boolean):void
		{
			_tlsRequired = tlsBool;
		}
		public function get tls():Boolean
		{
			return _tlsRequired;
		}
		
	}
}